/*****************************************************************************
 * Author:   Valient Gough <vgough@pobox.com>
 *
 *****************************************************************************
 * Copyright (c) 2002-2003, Valient Gough
 *
 * This library is free software; you can distribute it and/or modify it under
 * the terms of the GNU General Public License (GPL), as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GPL in the file COPYING for more
 * details.
 *
 */

#include "Cypher.h"

#include <string.h>

#include <openssl/blowfish.h>
#include <openssl/sha.h>
#include <openssl/rand.h>
#include <openssl/err.h>
#include <openssl/hmac.h>

#include "rlog.h"

using namespace std;

const int BLOCKSIZE = 64;

// TODO: less then about 6 rounds aren't enough for avalanche in all cases.
// How badly do we want to couple all bytes in each block?
// 2 rounds gives about 8 bytes of avalanche for a blocksize of 32 
// - similar to using BF_cbc_encrypt
// XXX: 2 rounds is the minimum, because we swap the array direction each
// round, and we want the last bytes to depend at least somewhat on the
// preceeding bytes.
const int CYPHER_ROUNDS = 2;

struct BlowfishKey
{
    BF_KEY key;
    unsigned char keyGenerator[SHA_DIGEST_LENGTH];
};


Cypher::Cypher()
{
}

Cypher::~Cypher()
{
}

/*
    create a blowfish key from the password.
    Use SHA to distribute entropy from the password into the key.

    This algorithm must remain constant for backward compatibility.
*/
void *Cypher::newKey(const char *password, int passwdLength)
{
    BlowfishKey *key = new BlowfishKey;
    SHA1( (const unsigned char*)password, passwdLength, key->keyGenerator );

    BF_set_key( &key->key, SHA_DIGEST_LENGTH, key->keyGenerator );

    return key;
}

/*
    Create a random key.
    We use the OpenSSL library to generate random bytes, then take the hash of
    those bytes to use as the key.

    This algorithm can change at any time without affecting backward
    compatibility.
*/
void *Cypher::newRandomKey()
{
    BlowfishKey *key = new BlowfishKey;
    unsigned char tmpBuf[ SHA_DIGEST_LENGTH ];
    if(RAND_bytes( tmpBuf, SHA_DIGEST_LENGTH ) == 0)
    {
	char errStr[120];
	unsigned long errVal = 0;
	if((errVal = ERR_get_error()) != 0)
	{
	    rDebug("openssl error: %s", ERR_error_string( errVal, errStr ));
	    delete key;
	    return 0;
	}
    }

    // take the hash, in case the RAND_bytes data isn't uniformly random..
    SHA1( tmpBuf, SHA_DIGEST_LENGTH, key->keyGenerator );

    BF_set_key( &key->key, SHA_DIGEST_LENGTH, key->keyGenerator );

    return key;
}

void Cypher::destroyKey(void *key)
{
    delete (BlowfishKey*)key;
}

void *Cypher::readKey(const unsigned char *data, void *key_)
{
    BlowfishKey *key = new BlowfishKey;
    
    if(!decodeName( (const char*)data, keySize(), (char*)key->keyGenerator, 
		    key_))
    {
	rWarning("readKey failed");
	return 0;
    }

    BF_set_key( &key->key, SHA_DIGEST_LENGTH, key->keyGenerator );

    return key;
}

void Cypher::writeKey(const void *key, unsigned char *data, void *key_)
{
    encodeName( (const char *)((BlowfishKey*)key)->keyGenerator, 
	    SHA_DIGEST_LENGTH, (char *)data, key_);
}

int Cypher::keySize() const
{
    return encodedNameLen( SHA_DIGEST_LENGTH );
}

int Cypher::blockSize() const
{
    return BLOCKSIZE;
}

int Cypher::encodedNameLen(int plaintextNameLen) const
{
    // we add 2 bytes for checksum info
    return plaintextNameLen + 2;
}

int Cypher::decodedNameLen(int encodedNameLen) const
{
    return encodedNameLen - 2;
}

static void setIVec( unsigned char ivec[8], unsigned int seed)
{
    unsigned int var1 = 0x060a4011 * seed;
    unsigned int var2 = 0x0221040d * (seed ^ 0xD3FEA11C);
	
    ivec[0] = (var1 >> 24) & 0xff;
    ivec[1] = (var2 >> 16) & 0xff;
    ivec[2] = (var1 >> 8 ) & 0xff;
    ivec[3] = (var2      ) & 0xff;
    ivec[4] = (var2 >> 24) & 0xff;
    ivec[5] = (var1 >> 16) & 0xff;
    ivec[6] = (var2 >> 8 ) & 0xff;
    ivec[7] = (var1      ) & 0xff;
}

static void reverse(const unsigned char *src, int size, unsigned char *dst)
{
    while(size)
	*dst++ = src[ --size ];
}

/*
    compute a 16-bit check value for the data using HMAC.
*/
static unsigned int _checksum( const unsigned char *key, int keyLength,
	const char *data, int dataLen)
{
    static const EVP_MD *evp_md = EVP_sha1();

    unsigned char md[EVP_MAX_MD_SIZE];
    unsigned int mdLen = EVP_MAX_MD_SIZE;

    unsigned char *mdigest = 
	HMAC( evp_md, key, keyLength, 
		(const unsigned char *)data, dataLen, md, &mdLen);

    rAssert(mdigest != 0);

    // chop this down to a 16bit value..
    unsigned char h1 = 0;
    unsigned char h2 = 0;
    for(unsigned int i=0; i<(mdLen-1); i+=2)
    {
	h1 ^= (unsigned char)(mdigest[i]);
	h2 ^= (unsigned char)(mdigest[i+1]);
    }

    return ((((unsigned int)h1) << 8) | ((unsigned int)h2));
}

void Cypher::encodeName(const char *src, int len, char *dst, void *key_)
{
    BlowfishKey *key = (BlowfishKey*)key_;

    int ivecNum = 0;
    unsigned char ivec[8];

    // cfb stuff
    // add checksum, making filenames much more randomized..  Also provides a
    // way to check that decoding the name was successful.  Downside is that
    // having the checksum available allows an attacker information about the
    // filename without knowing the decryption key..  So, encrypt the
    // checksum..
    unsigned int checksum = _checksum( key->keyGenerator, SHA_DIGEST_LENGTH,
	    src, len );

    //rWarning("encoding data len %i, checksum = %i", len, checksum);

    setIVec( ivec, checksum );
    ivecNum = 0;

    BF_cfb64_encrypt((unsigned char*)src, (unsigned char *)dst, 
	             len, &key->key, ivec, &ivecNum, BF_ENCRYPT);

    dst[len] = (checksum >> 8) & 0xff;
    dst[len+1] = (checksum   ) & 0xff;
}

bool Cypher::decodeName(const char *src, int len, char *dst, void *key_)
{
    if(len <= 2)
	return false;

    BlowfishKey *key = (BlowfishKey*)key_;
    unsigned char ivec[8];
    int ivecNum = 0;

    // cfb decrypt
    unsigned int checksum = ((unsigned int)((unsigned char)src[len-2])) << 8
	                | ((unsigned int)((unsigned char)src[len-1]));
	
    setIVec( ivec, checksum );
    ivecNum = 0;

    BF_cfb64_encrypt((unsigned char*)src, (unsigned char*)dst, 
	             len-2, &key->key, ivec, &ivecNum, BF_DECRYPT);

    // check for success
    unsigned int checksum2 = _checksum( key->keyGenerator, SHA_DIGEST_LENGTH,
	    dst, len-2 );
    if(checksum2 != checksum)
    {
	rDebug("checksum mismatch: expected %u, got %u", checksum, checksum2);
	rDebug("on decode of %i bytes", len-2);
    }

    return (checksum == checksum2);
}

/* Some possible encoding options for a block of data:
   - encode in block-cypher mode.
     Advantages:
        - change in one byte affects many other bytes, making it difficult to
	  change cyphertext without being noticed.
	- does not lead to XOR analysis attacks
     Disadvantages:
        - must always work in blocks of bytes.  Files must be padded to the
	  block size, and then the exact size stored elsewhere.
   - encode in pseudo one-time-pad mode (counter mode).
     Advantages:
        - easy to randomly seek to position in cyphertext, as each byte is
	  encoded separately.
     Disadvantages:
        - can lead to XOR based attacks if cyphertext is sent more then once
	  and has been modified.  
	- data can be changed without knowing the encryption key
   - encode in pseudo-block mode using stream cypher
     Advantages:
        - same advantages as block-cypher mode
	- blocks can be arbitrarily sized, so file does not need to be padded.
*/
static bool codePartialBlock(char *buf, int size, unsigned long blockNum, 
	void *key_, int enc, int rounds)
{
    BlowfishKey *key = (BlowfishKey*)key_;

    rAssert( size < BLOCKSIZE );

    unsigned char encBuf[ BLOCKSIZE ];
    unsigned char ivec[8];
    for(int round = rounds; round; --round)
    {
	int ivecNum = 0;
	setIVec( ivec, blockNum );

	// buf --> encBuf
	BF_cfb64_encrypt( (unsigned char *)buf, encBuf, size, 
		&key->key, ivec, &ivecNum, enc);

	if( round > 1 )
	{
	    // reverse, so that we can do a second pass the other direction
	    // through the bytes.  That way we make the front of the block
	    // dependent on the other end of the block.
	    //   buf <-- encBuf 
	    reverse( encBuf, size, (unsigned char *)buf );
	}
    }

    if(rounds)
	memcpy( buf, encBuf, size ); // buf <-- encBuf

    return true;
}

static bool codeFullBlock(char *buf, int numBlocks, unsigned long blockNum, 
	void *key_, int enc )
{
    BlowfishKey *key = (BlowfishKey*)key_;

    unsigned char encBuf[ BLOCKSIZE ];
    unsigned char ivec[8];

    for(int block=0; block<numBlocks; ++block)
    {
	setIVec( ivec, blockNum );

	BF_cbc_encrypt( (unsigned char *)buf, encBuf, BLOCKSIZE, 
		&key->key, ivec, enc);

	memcpy( buf, encBuf, BLOCKSIZE ); // buf <-- encBuf

	++blockNum;
	buf += BLOCKSIZE;
    }

    return true;
}

bool Cypher::encodeBlock(char *buf, int size, unsigned long blockNum,
	void *_key)
{
    int numBlocks = size / BLOCKSIZE;
    int remainder = size % BLOCKSIZE;

    bool ok = true;
    if(numBlocks)
	ok = codeFullBlock( buf, numBlocks, blockNum, _key, BF_ENCRYPT );

    if(remainder)
	ok &= codePartialBlock( buf+numBlocks*BLOCKSIZE, remainder, 
		blockNum+numBlocks, _key, BF_ENCRYPT, CYPHER_ROUNDS );

    return ok;
}

bool Cypher::decodeBlock(char *buf, int size, unsigned long blockNum,
	void *_key)
{
    int numBlocks = size / BLOCKSIZE;
    int remainder = size % BLOCKSIZE;

    bool ok = true;
    if(numBlocks)
	ok = codeFullBlock( buf, numBlocks, blockNum, _key, BF_DECRYPT );

    if(remainder)
	ok &= codePartialBlock( buf+numBlocks*BLOCKSIZE, remainder, 
		blockNum+numBlocks, _key, BF_DECRYPT, CYPHER_ROUNDS );

    return ok;
}

