/*****************************************************************************
 * Author:   Valient Gough <vgough@pobox.com>
 *
 *****************************************************************************
 * Copyright (c) 2003, Valient Gough
 *
 * This library is free software; you can distribute it and/or modify it under
 * the terms of the GNU General Public License (GPL), as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GPL in the file COPYING for more
 * details.
 *
 */

#include "FileInfo.h"

#include "base64.h"
#include "Cypher.h"
#include "rlog.h"

using namespace std;


FileInfo::FileInfo(const char *filename, Cypher *cypher, void *key)
    : LinkedOwner()
{
    data = new FileInfo::Data;
    data->name = filename;
    data->cypherName = encodeName(filename, cypher, key);
}

FileInfo::FileInfo(const FileInfo &src)
    : LinkedOwner()
{
    shareOwnership( &src );
    data = src.data;
}

FileInfo::~FileInfo()
{
    if(data && dropOwnership())
    {
	delete data;
	data = 0;
    }
}

FileInfo & FileInfo::operator = (const FileInfo &src)
{
    if(this != &src)
    {
	if(data && dropOwnership())
	{
	    delete data;
	    data = 0;
	}

	shareOwnership( &src );
	data = src.data;
    }

    return *this;
}

/*
    Encode filename.
    We don't encode '/', or '.' or '..'
*/
string FileInfo::encodeName(const char *plaintextName,
	Cypher *cypher, void *key)
{
    string output;

    char tmpBuf[32];

    while(*plaintextName)
    {
	if(*plaintextName == '/')
	{
	    if(!output.empty()) // don't start the string with '/'
		output += '/';
	    while(*plaintextName == '/')
	       	++plaintextName;
	} else
	{
	    char *next = strchr( plaintextName, '/' );
	    int len = next ? next - plaintextName : strlen( plaintextName );

	    if(len <=2 && *plaintextName == '.')
	    {
		if(len == 1)
		{
		    output += ".";
		    plaintextName += len;
		    continue;
		} else
		if(len == 2 && plaintextName[1] == '.')
		{
		    output += "..";
		    plaintextName += len;
		    continue;
		}
	    }

	    int encLen = cypher->encodedNameLen( len );
	    int encLenB64 = B256ToB64Bytes( encLen );
	    char *encBuf = tmpBuf;
	    if(encLenB64 >= (int)sizeof(tmpBuf))
	    {
		encBuf = new char[encLenB64+1];
		memset(encBuf, 0, encLenB64+1);
	    } else
		memset(encBuf, 0, sizeof(tmpBuf));

	    cypher->encodeName( plaintextName, len, encBuf, key );
	    plaintextName += len;

	    changeBase2Inline( (unsigned char*)encBuf, encLen, 8, 6, true );
	    B64ToAscii( (unsigned char*)encBuf, encLenB64 );
	    rAssert(encBuf[ encLenB64 ] == '\0');

	    output += encBuf;
	    if(encBuf != tmpBuf)
		delete[] encBuf;
	}
    }

    return output;
}

string FileInfo::decodeName(const char *cypherName,
	Cypher *cypher, void *key)
{
    string output;

    unsigned char transBuf[32];  // translation buffer
    unsigned char tmpBuf[32];

    while(*cypherName)
    {
	if(*cypherName == '/')
	{
	    output += '/';

	    while(*cypherName == '/')
	       	++cypherName;
	} else
	{
	    char *next = strchr( cypherName, '/' );
	    int len = next ? next - cypherName : strlen( cypherName );
	    
	    if(len <=2 && *cypherName == '.')
	    {
		if(len == 1)
		{
		    output += ".";
		    cypherName += len;
		    continue;
		} else
		if(len == 2 && cypherName[1] == '.')
		{
		    output += "..";
		    cypherName += len;
		    continue;
		}
	    }

	    // translate to 8-bit data
	    unsigned char *tBuf = transBuf;
	    if(len >= (int)sizeof(transBuf))
		tBuf = new unsigned char[ len+1 ];

	    AsciiToB64( tBuf, (unsigned char*)cypherName, len );
	    changeBase2Inline( tBuf, len, 6, 8, false);

	    // decode from tBuf -> tmpBuf
	    int plainLen = cypher->decodedNameLen( B64ToB256Bytes( len ) );

	    unsigned char *plainBuf = tmpBuf;
	    if(plainLen >= (int)sizeof(tmpBuf))
		plainBuf = new unsigned char[plainLen+1];

	    if(!cypher->decodeName( (const char*)tBuf, 
			B64ToB256Bytes( len ), (char*)plainBuf, key))
	    {
		rDebug("decode failed for %s, length %i", 
			cypherName, (int)len);
		return string();
	    }
	    plainBuf[ plainLen ] = '\0';
	    cypherName += len;

	    output += (char*)plainBuf;
	    if(plainBuf != tmpBuf)
		delete[] plainBuf;
	    if(tBuf != transBuf)
		delete[] tBuf;
	}
    }

    return output;
}

string FileInfo::name() const
{
    return data->name;
}

string FileInfo::cypherName() const
{
    return data->cypherName;
}

