/*****************************************************************************
 * Author:   Valient Gough <vgough@pobox.com>
 *
 *****************************************************************************
 * Copyright (c) 2003, Valient Gough
 * 
 * This program is free software; you can distribute it and/or modify it under 
 * the terms of the GNU General Public License (GPL), as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 */
                             
#ifdef linux
/* For pread()/pwrite() */
#define _XOPEN_SOURCE 500
#endif

#include "encfs.h"

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/statfs.h>

#include <string>
#include <map>

#include "Cypher.h"
#include "FileInfo.h"
#include "rlog.h"
#include "MemoryPool.h"

#ifndef MIN
#define MIN(a,b) (((a)<(b)) ? (a): (b))
#endif

using namespace std;

extern Cypher *gCypher;
extern int gBlockSize;
extern void *gVolumeKey;
extern string gRootDir; // always ends in '/'

pthread_mutex_t gFDStatChange = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t gMPoolMutex = PTHREAD_MUTEX_INITIALIZER;
MemoryPool *gMPool =0;

struct OpenFDInfo
{
    int fd;
    int oldfd;
    bool canWrite;
    int refCount;
};

pthread_mutex_t gFDMapMutex = PTHREAD_MUTEX_INITIALIZER;
static map<string, OpenFDInfo> gFDMap;
static char * allocBlock( )
{
    pthread_mutex_lock( &gMPoolMutex );
    if(!gMPool)
	gMPool = new MemoryPool( gBlockSize );

    char *result = (char*)gMPool->allocate();
    pthread_mutex_unlock( &gMPoolMutex );

    return result;
}

static void releaseBlock( char *block )
{
    pthread_mutex_lock( &gMPoolMutex );

    gMPool->release( (void*)block );

    pthread_mutex_unlock( &gMPoolMutex );
}

static ssize_t readBlock( int fd, char *blockPtr, unsigned long blockNum )
{
    //rDebug("reading block %i", (int)blockNum);

    ssize_t readSize = pread( fd, blockPtr, gBlockSize,
	                      blockNum * gBlockSize );

    if(readSize >= 0)
    {
	if(!gCypher->decodeBlock( blockPtr, (int)readSize, blockNum,
		gVolumeKey ))
	{
	    rDebug("decodeBlock failed for block %i, size %i",
		    (int)blockNum, (int)readSize);
	    readSize = -1; // error
	}
    } else
    {
	int eno = errno;
	rDebug("pread failed at block %i for %i bytes: %s", 
		(int)blockNum, gBlockSize, strerror( eno ));
    }

    return readSize;
}


static bool writeBlock( int fd, char *blockPtr, unsigned long blockNum, 
    ssize_t size )
{
    //rDebug("writeBlock for block %i, size %i", (int)blockNum, (int)size);
    /*
    if(size < gBlockSize)
	rDebug("writing partial block num %i, size %i", 
		(int)blockNum, (int)size);
     */
    if( gCypher->encodeBlock( blockPtr, size, blockNum, gVolumeKey ) )
    {
	ssize_t writeSize = pwrite( fd, blockPtr, size, 
		blockNum * gBlockSize );

	if( writeSize == size )
	{
	    return true;
	}
    }

    rDebug("writeBlock failed");
    return false;
}

static string cypherName( const char *path )
{
    string result = gRootDir;
    result += FileInfo::encodeName( path, gCypher, gVolumeKey );
    return result;
}

// workaround for opening a file for write when permissions don't allow.
// Since the kernel has already checked permissions, we can assume it is ok to
// provide access.. So force it by changing permissions temporarily.
// We use a mutex around this so that there won't be a race condition with
// other calls to lstat picking up the wrong permissions..
static int open_readonly_workaround( const string &cypherPath, int flags )
{
    int fd = -1;
    pthread_mutex_lock( &gFDStatChange );
    struct stat stbuf;
    if(lstat( cypherPath.c_str(), &stbuf ) != -1)
    {
	// make sure user has read/write permission..
	chmod( cypherPath.c_str(), stbuf.st_mode | 0600 );
	fd = open( cypherPath.c_str(), flags );
	chmod( cypherPath.c_str(), stbuf.st_mode );
    } else
    {
	// can't even stat the file... forget it.
	rDebug("can't stat file %s", cypherPath.c_str());
    }

    pthread_mutex_unlock( &gFDStatChange );

    return fd;
}

// We shouldn't have to support all possible open flags, so untaint the flags
// argument by only taking ones we understand/accept.
// - Since the kernel has already done permission tests before calling us, we
// shouldn't have to worry about access control.
// - Basically we just need to distinguish between read and write flags.
static int openHandle( const string &cypherPath, int flags )
{
    //rDebug("openHandle: %s, %i", cypherPath.c_str(), flags);

    int finalFlags = flags & (O_RDWR | O_RDONLY | O_WRONLY);

    if( flags & O_WRONLY )
    {
	// always allow reads..
	finalFlags &= ~O_WRONLY;
	finalFlags |= O_RDWR;
    }
   
    int result = 0;
    OpenFDInfo info;

    pthread_mutex_lock( &gFDMapMutex );
    map<string, OpenFDInfo>::iterator it = gFDMap.find( cypherPath );

    // get existing info structure, or initialize a new one
    if(it != gFDMap.end())
    {
	++it->second.refCount;
	info = it->second;
    } else
    {
	info.fd = info.oldfd = -1;
	info.canWrite = false;
	info.refCount = 1;
    }
    pthread_mutex_unlock( &gFDMapMutex );

    /*
	If we already have a file descriptor, and we're allowed to write on
	it, or they aren't asking for a writable descriptor, then use it..
     */
    if((info.fd >= 0) && 
	    (info.canWrite || 
	     (finalFlags & (O_RDONLY | O_WRONLY | O_RDWR) == O_RDONLY)))
    {
	// no need to re-open, use what we've got
	result = info.fd;
    } else
    {
	int fd = open( cypherPath.c_str(), finalFlags );

	if((fd == -1) && (errno == EACCES))
	    fd = open_readonly_workaround( cypherPath, finalFlags );

	if(fd != -1)
	{
	    // the old fd might still be in use, so we might not be able to
	    // get rid of it just yet..
	    info.oldfd = info.fd;
	    info.fd = fd;
	    info.canWrite = (finalFlags & O_RDWR);
	
	    result = fd;
   
	    // update global table
	    pthread_mutex_lock( &gFDMapMutex );
	    it = gFDMap.find( cypherPath );
	    if((it != gFDMap.end()) && (it->second.oldfd != -1))
	    {
		// oops, race condition - someone re-opened the file before we
		// could get to it..  Use what's there..
		close(fd);
		result = it->second.fd;
	    } else
	    {
		gFDMap[cypherPath] = info;
	    }
	    pthread_mutex_unlock( &gFDMapMutex );
	} else
	    result = -errno;
    }

    if(result < 0)
    {
	rDebug("file %s open failure: %i", cypherPath.c_str(), -result);
    }
    return result;
}

static void closeHandle( const string &cypherPath )
{
    //rDebug("closeHandle: %s", cypherPath.c_str());

    pthread_mutex_lock( &gFDMapMutex );

    map<string, OpenFDInfo>::iterator it = gFDMap.find( cypherPath );
    if(it != gFDMap.end())
    {
	int count = --(it->second.refCount);
	rAssert( count >= 0 );
	if(!count)
	{
	    if( it->second.fd != -1 )
	       	close( it->second.fd );
	    if( it->second.oldfd != -1 )
		close( it->second.oldfd );

	    gFDMap.erase( it );
	}
    }
    
    pthread_mutex_unlock( &gFDMapMutex );
}

// Similar to closeHandle, except that it indicates that there are no more
// users of this file, so all open references can be destroyed..
static void releaseHandle( const string &cypherPath )
{
    //rDebug("releaseHandle: %s", cypherPath.c_str());

    pthread_mutex_lock( &gFDMapMutex );

    map<string, OpenFDInfo>::iterator it = gFDMap.find( cypherPath );
    if(it != gFDMap.end())
    {
	if( it->second.fd != -1 )
	    close( it->second.fd );
	if( it->second.oldfd != -1 )
	    close( it->second.oldfd );

	gFDMap.erase( it );
    }
    
    pthread_mutex_unlock( &gFDMapMutex );
}

int encfs_getattr(const char *path, struct stat *stbuf)
{
    // lookup file info
    string cyName = cypherName( path );

    //rDebug("lstat of %s", (const char *)cyName);
    const char *cname = cyName.c_str();
    pthread_mutex_lock( &gFDStatChange );
    int res = lstat( cname, stbuf );
    pthread_mutex_unlock( &gFDStatChange );
    int eno = errno;
    if(res < 0)
    {
	if(!eno)
	{
	    rWarning("errno not set after lstat error!");
	    eno = ENOENT;
	}
	//rDebug("lstat error on %s: %s (%i)", cname, strerror(eno), eno);
        return -eno;
    }

    return 0;
}

int encfs_getdir(const char *path, fuse_dirh_t h, fuse_dirfil_t filler)
{
    int res = 0;

    string cyName = cypherName( path );

    //rDebug("getdir of %s", (const char *)cyName);
    DIR *dp = ::opendir( cyName.c_str() );
    if(!dp)
    {
       	int eno = errno;
	rDebug("opendir error %s", strerror(eno));
        return -eno;
    }

    struct dirent *de;
    while((de = readdir(dp)) != NULL)
    {
	if( de->d_name[0] != '.' )
	{
	    //cout << "decoding filename " << de->d_name << "\n";
	    // decode name
	    string name = FileInfo::decodeName( de->d_name, gCypher,
		                                 gVolumeKey );

	    if(!name.empty())
	    {
		//cout << "decoded name = " << name << "\n";
		// XXX: no Unicode support?!
		res = filler( h, name.c_str(), de->d_type );
	    } 
	} else
	{
	    if( (de->d_name[1] == '\0') || 
		      (de->d_name[1] == '.' && de->d_name[2] == '\0'))
	    {
		res = filler( h, de->d_name, de->d_type );
	    }
	}

	if(res != 0)
   	    break;
    }

    closedir(dp);
    return res;
}

int encfs_mknod(const char *path, mode_t mode, dev_t rdev)
{
    string cyName = cypherName( path );

    //rDebug("mknod on %s, mode %i", (const char *)cyName.c_str(), (int)mode);

    int res = mknod( cyName.c_str(), mode, rdev );
    if(res == -1)
    {
	int eno = errno;
	rDebug("mknod error: %s", strerror(eno));
	return -eno;
    }

    return 0;
}

int encfs_mkdir(const char *path, mode_t mode)
{
    string cyName = cypherName( path );

    //rDebug("mkdir %s, mode %i", (const char *)cyName, mode);
    int res = mkdir( cyName.c_str(), mode );
    if(res == -1)
    {
       	int eno = errno;
	rDebug("mkdir error: %s", strerror(eno));
        return -eno;
    }

    return 0;
}

int encfs_unlink(const char *path)
{
    string cyName = cypherName( path );

    //rDebug("unlink %s", (const char *)cyName);
    int res = unlink( cyName.c_str() );
    if(res == -1)
    {
       	int eno = errno;
	rDebug("unlink error: %s", strerror(eno));
        return -eno;
    }

    return 0;
}

int encfs_rmdir(const char *path)
{
    string cyName = cypherName( path );

    //rDebug("rmdir %s", (const char *)cyName);
    int res = rmdir( cyName.c_str() );
    if(res == -1)
    {
       	int eno = errno;
	rDebug("unlink error: %s", strerror(eno));
        return -eno;
    }

    return 0;
}

static bool isOrphaned(const string &cname)
{
    (void)cname;
    // TODO: Check if the file is still open (we have a file descriptor in
    // memory) but the file itself has been removed.  We don't want to allow
    // recreating the file in this case, otherwise - boom, confusion!
    return false;
}

int encfs_readlink(const char *path, char *buf, size_t size)
{
    string cyName = cypherName( path );

    int res = readlink( cyName.c_str(), buf, size - 1 );

    if(res == -1)
        return -errno;

    buf[res] = '\0'; // null terminate
    string decodedName = FileInfo::decodeName(buf, gCypher, gVolumeKey);

    if(!decodedName.empty())
    {
	strcpy(buf, decodedName.c_str());

	return 0;
    } else
    {
	rWarning("Error decoding link");
	return -1;
    }
}

int encfs_symlink(const char *from, const char *to)
{
    // symlink doesn't need the regular path processing..
    string fromCName = FileInfo::encodeName( from, gCypher, gVolumeKey );
    string toCName = cypherName( to );

    if(isOrphaned(to))
	return -EINVAL;

    int res = symlink( fromCName.c_str(), toCName.c_str() );
    if(res == -1)
        return -errno;

    return 0;
}

int encfs_link(const char *from, const char *to)
{
    string fromCName = cypherName( from );
    string toCName = cypherName( to );

    if(isOrphaned(to))
	return -EINVAL;

    int res = link( fromCName.c_str() , toCName.c_str() );
    if(res == -1)
        return -errno;

    return 0;
}


int encfs_rename(const char *from, const char *to)
{
    string fromCName = cypherName( from );
    string toCName = cypherName( to );

    if(isOrphaned(to))
	return -EINVAL;

    int res = rename( fromCName.c_str(), toCName.c_str() );
    if(res == -1)
        return -errno;

    return 0;
}

int encfs_chmod(const char *path, mode_t mode)
{
    string cyName = cypherName( path );

    //rDebug("chmod %s, mode %i", cyName.c_str(), mode);
    int res = chmod( cyName.c_str(), mode);
    if(res == -1)
    {
       	int eno = errno;
	rDebug("chmod error: %s", strerror(eno));
        return -eno;
    }

    return 0;
}

int encfs_chown(const char *path, uid_t uid, gid_t gid)
{
    string cyName = cypherName( path );

    int res = lchown( cyName.c_str(), uid, gid);
    if(res == -1)
        return -errno;

    return 0;
}

int encfs_truncate(const char *path, off_t size)
{
    string cyName = cypherName( path );

    // if we are truncating within a block boundary, we have to read what is
    // there and write the partial block before truncating..
    int partialBlock = size % gBlockSize;
    if(partialBlock)
    {
	int fd = openHandle( cyName, O_RDWR );
	if(fd < 0)
	    return fd;

	unsigned long blockNum = size / gBlockSize;
	char *blockPtr = allocBlock( );

	// read a block and write back part of it..
	int rdSz = readBlock( fd, blockPtr, blockNum );
	int wrSz = writeBlock( fd, blockPtr, blockNum, partialBlock );
	if(!rdSz || !wrSz)
	    rWarning("truncate: read %i bytes, wrote %i bytes", 
		    rdSz, wrSz);

	releaseBlock( blockPtr );
	closeHandle( cyName );
    }

    if (truncate( cyName.c_str(), size) == -1)
    {
	int eno = errno;
	rDebug("truncate failed for %s size %i, error %s", cyName.c_str(), 
		(int)size, strerror( eno ));
	return -errno;
    } else
	return 0;
}

int encfs_utime(const char *path, struct utimbuf *buf)
{
    string cyName = cypherName( path );

    int res = utime( cyName.c_str(), buf);
    if(res == -1)
    {
	int eno = errno;
	rDebug("utime error: %s", strerror(eno));
        return -eno;
    }

    return 0;
}


int encfs_open(const char *path, int flags)
{
    string cyName = cypherName( path );

    //rDebug("encfs_open of %s, flags %i", path, flags);
    int res = openHandle( cyName, flags );
    if(res < 0)
    {
	int eno = -res;
	rDebug("open error: %s on file %s, flags = %i", strerror(eno), 
		cyName.c_str(), flags);
        return res;
    }

    //closeHandle(res);
    return 0; // 0 indicates no error
}

int encfs_release(const char *path, int flags)
{
    (void)flags;

    //rDebug("encfs_release of %s, flags %i", path, flags);
    string cyName = cypherName( path );

    releaseHandle( cyName );

    return 0;
}

int encfs_read(const char *path, char *buf, size_t size, off_t offset)
{
    //rDebug("requested %i bytes at %i from %s", (int)size, (int)offset, path);
    string cyName = cypherName( path );

    int fd = openHandle( cyName, O_RDONLY);
    if(fd < 0)
        return fd;

    unsigned long blockNum = offset / gBlockSize;
    int partialOffset = offset % gBlockSize;
    size_t origSize = size;

    // use one working block
    char *blockPtr = allocBlock( );
    while(size)
    {
	// TODO: check for failure in readBlock
	int readSize = readBlock( fd, blockPtr, blockNum );
	if(readSize <= 0)
	    break; // attempted read-past-eof? 

	if(readSize <= partialOffset)
	    break; // didn't get enough bytes
	int cpySize = MIN( (size_t)(readSize - partialOffset), size );

	//rDebug("copy %i bytes from block %i, offset %i",
	//	cpySize, (int)blockNum, partialOffset);
	memcpy( buf, blockPtr + partialOffset, cpySize );

       	size -= cpySize;
	buf += cpySize;
	++blockNum;
	partialOffset = 0;
	
	if(readSize < gBlockSize)
	    break;
    }
	
    releaseBlock( blockPtr );
    closeHandle( cyName );

    return origSize - size;
}

static off_t getFileSize( const char *path )
{
    struct stat stbuf;
    // we don't need to hold gFDStatChange mutex because it is only for
    // ownership stat info, which we won't use..
    if(lstat( path, &stbuf ) == 0)
	return stbuf.st_size;
    else
	return -1;
}

static void padFile(int fd, int oldSize, int newSize)
{
    //rDebug("pading file from size %i to %i", oldSize, newSize);

    unsigned long blockNum = oldSize / gBlockSize;
    unsigned long lastBlock = newSize / gBlockSize;

    char *blockPtr = allocBlock( );
    memset( blockPtr, 0, gBlockSize );
    if(blockNum == lastBlock)
    {
	int toWrite = newSize % gBlockSize;
	//rDebug("enlarging block %i to %i bytes", (int)blockNum,
	//	(int)toWrite);
	readBlock( fd, blockPtr, blockNum );
	writeBlock( fd, blockPtr, blockNum, toWrite );
    } else
    {
	if(blockNum % gBlockSize != 0)
	{
	    // enlarge first block to full size
	    readBlock( fd, blockPtr, blockNum );
	    writeBlock( fd, blockPtr, blockNum, gBlockSize );
	}

	// fill in the rest with zeros
	for(++blockNum; blockNum != lastBlock; ++blockNum)
	{
	    memset( blockPtr, 0, gBlockSize );
	    writeBlock( fd, blockPtr, blockNum, gBlockSize );
	}

	// write last block, which may be a partial block
	int toWrite = newSize % gBlockSize;
	memset( blockPtr, 0, gBlockSize );
	writeBlock( fd, blockPtr, blockNum, toWrite );
    }
    releaseBlock( blockPtr );
}

int encfs_fsync(const char *path, int dataSync)
{
    string cyName = cypherName( path );
    int res = 0;

    int fd = openHandle( cyName, O_RDWR );
    if(fd < 0)
    {
	rDebug("error getting RDWR descriptor for %s", cyName.c_str());
    } else
    {
	if(dataSync == 0)
	    res = fsync( fd );
	else
	    res = fdatasync( fd );
    }

    closeHandle( cyName );

    return res;
}

int encfs_write(const char *path, const char *buf, size_t size,
                     off_t offset)
{
    //rDebug("requested write of %i bytes at offset %i to %s",
    //    (int)size, (int)offset, path);
    string cyName = cypherName( path );

    int fd = openHandle( cyName, O_RDWR);
    if(fd < 0)
    {
	int eno = -fd;
	rDebug("error opening file %s for write: %s", cyName.c_str(), 
		strerror(eno));
        return fd;
    }
    
    unsigned long blockNum = offset / gBlockSize;
    int partialOffset = offset % gBlockSize;
    size_t origSize = size;
   
    // check current size - we may need to extend the file first..
    // TODO: good place for optimization!
    off_t fileSize = getFileSize( cyName.c_str() );
    if(fileSize < 0)
	fileSize = 0;

    if(offset > fileSize)
    {
	// pad file with 0's.  We can't let the underlying filesystem do this,
	// because it would pad with unencrypted 0's, so it wouldn't look right
	// after decrypting!
	padFile( fd, fileSize, offset );
    }

    char *blockPtr = allocBlock( );
    while(size)
    {
	int cpySize = MIN( (size_t)(gBlockSize - partialOffset), size );
	
	//rDebug("writing %i bytes in block %i, at offset %i",
	//	(int)cpySize, (int)blockNum, (int)partialOffset);

	// TODO: this isn't necessary if we're at the end of the file..
	ssize_t curBlockSize = gBlockSize;
	if(cpySize != gBlockSize)
	{
	    //rDebug("overwriting partial block, num %i", (int)blockNum);
	    curBlockSize = readBlock( fd, blockPtr, blockNum );
	    //rDebug("read %i bytes", curBlockSize);
	    if(curBlockSize < cpySize + partialOffset)
		curBlockSize = cpySize + partialOffset;
	}

	memcpy( blockPtr + partialOffset, buf, cpySize );

	// TODO: check for failure in readBlock / writeBlock
	writeBlock( fd, blockPtr, blockNum, curBlockSize );

       	size -= cpySize;
	buf += cpySize;
	++blockNum;
	partialOffset = 0;
    }

    releaseBlock( blockPtr );
    closeHandle( cyName );
    return origSize - size;
}

int encfs_statfs(struct fuse_statfs *fst)
{
    struct statfs st;
    // TODO: replace with root mount point
    int rv = statfs( gRootDir.c_str(),&st );
    if(!rv) 
    {
    	fst->block_size  = st.f_bsize;
    	fst->blocks      = st.f_blocks;
    	fst->blocks_free = st.f_bavail;
    	fst->files       = st.f_files;
    	fst->files_free  = st.f_ffree;
    	fst->namelen     = st.f_namelen;
    }
    return rv;
}

