/*****************************************************************************
 * Author:   Valient Gough <vgough@pobox.com>
 *
 *****************************************************************************
 * Copyright (c) 2003, Valient Gough
 *
 * This library is free software; you can distribute it and/or modify it under
 * the terms of the GNU General Public License (GPL), as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GPL in the file COPYING for more
 * details.
 *
 */

#include "encfs.h"
#include "config.h"

#include <openssl/ssl.h>

#include <iostream>
#include <string>

#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>

#include "FileInfo.h"
#include "Cypher.h"
#include "readpassphrase.h"
#include "rlog.h"

using namespace std;

const char V2KeyName[] = ".encfs3";

Cypher *gCypher =0;
int gBlockSize =0;
string gRootDir;
void *gVolumeKey =0;

const int MaxHeaderSize = 32;

static bool readVolumeKey(const string &filePath)
{
    // read it!
    int headerSize = gCypher->keySize();

    char passBuf[64];
    char *res = readpassphrase( "Password: ", passBuf, sizeof(passBuf)-1, 
	    RPP_ECHO_OFF);

    if(!res)
    {
	rWarning("password required to mount volume");
	return false;
    }

    void *masterKey = gCypher->newKey( passBuf, strlen(passBuf) );
    memset(passBuf, 0, sizeof(passBuf));

    int vkeyFD = open( filePath.c_str(), O_RDONLY );
    if( vkeyFD >= 0 )
    {
	assert(headerSize <= MaxHeaderSize);
	char encodeKey[ MaxHeaderSize ];

	read( vkeyFD, encodeKey, headerSize );
	close( vkeyFD );

	// decode volume key
	gVolumeKey = gCypher->readKey( 
		(unsigned char *)encodeKey,
		masterKey);

	if(!gVolumeKey)
	{
	    gCypher->destroyKey(masterKey);
	    cerr << "Unable to decode volume key - incorrect password?\n";
	    return false;
	}
    } else
    {
	// failed to open - we know it exists, so there must be a
	// permission problem.
	gCypher->destroyKey(masterKey);
	rWarning("Unable to open volume key file: %s", filePath.c_str());
	return false;
    }

    gCypher->destroyKey(masterKey);
    return true;
}

static bool createVolumeKey( const string &mkeyPath )
{
    void *masterKey = 0;
    while(1)
    {
	// creating new volume key.. should check that is what the user is
	// expecting...
	cout << "Volume key not found, creating new encrypted volume.\n";

	char passBuf[64];
	char *res = readpassphrase( "Password: ", passBuf, 
		sizeof(passBuf)-1, RPP_ECHO_OFF);

	char passBuf2[64];
	res = readpassphrase( "Verify: ", passBuf2, 
		sizeof(passBuf2)-1, RPP_ECHO_OFF);

	if(!strcmp(passBuf, passBuf2))
	{
	    masterKey = gCypher->newKey( passBuf, strlen(passBuf) );
	    memset(passBuf, 0, sizeof(passBuf));
	    memset(passBuf2, 0, sizeof(passBuf2));
	    break;
	} else
	{
	    cerr << "Password was not typed correctly, please try again\n";
	}
    }

    gVolumeKey = gCypher->newRandomKey();
    if(!gVolumeKey)
    {
	rWarning("Unable to generate new key!");
	return false;
    }

    int vkeyFD = open( mkeyPath.c_str(), O_RDWR | O_CREAT, 0640 );
    if( vkeyFD >= 0 )
    {
	int headerSize = gCypher->keySize();
	assert( headerSize <= MaxHeaderSize );
	unsigned char encodedKey[ MaxHeaderSize ];

	gCypher->writeKey( gVolumeKey, encodedKey, masterKey);

	write( vkeyFD, (char*)encodedKey, headerSize );
	close( vkeyFD );

	// try reading it back in!
	void *volumeKey = gCypher->readKey( encodedKey, masterKey );
	if(!volumeKey)
	{
	    rWarning("Internal error - unable to parse newly created key!");
	}
	gCypher->destroyKey( volumeKey );
    } else
    {
	cerr << "Permission denied: unable to open volume key file\n";
	gCypher->destroyKey( masterKey );
	return false;
    }

    gCypher->destroyKey( masterKey );
    return true;
}

static bool fileExists(const string &fileName)
{
    struct stat buf;
    if(!lstat( fileName.c_str(), &buf ))
    {
	return true;
    } else
    {
	// XXX  show perror
	return false;
    }
}

static bool isDirectory( const string &fileName)
{
    struct stat buf;
    if(!lstat( fileName.c_str(), &buf ))
    {
	return S_ISDIR( buf.st_mode );
    } else
    {
	// XXX show perror
	return false;
    }
}

/*  Ask the user if we can create this directory.  If they say ok, then do it..
*/
static bool userAllowCreate( const string &fileName )
{
    cerr << "Directory \"" << fileName << "\" does not exist, create (y,n)?";
    string answer;
    cin >> answer;

    if(!answer.empty() && toupper(answer[0]) == 'Y')
    {
	int result = mkdir( fileName.c_str() , 0700 );
	if(result < 0)
	{
	    perror("Unable to create directory: ");
	    return false;
	} else
	{
	    return true;
	}
    } else
    {
	cerr << "Directory not created\n";
	return false;
    }
}

static bool isAbsolute( const string &fileName)
{
    if(fileName.length() > 1 &&
	    fileName[0] == '/')
    {
	return true;
    } else
	return false;
}

static bool initVolumeKey()
{
    // load the master key file from the root directory
    string keyName = gRootDir + V2KeyName;

    if(fileExists( keyName ))
    {
	return readVolumeKey( keyName );
    } else
    {
	if(isDirectory( gRootDir ))
	{
	    return createVolumeKey( keyName );
	} else
	{
	    rWarning("No volume in root directory, and unable to create one.");
	    return false;
	}
    }

    return true;
}

void usage(const char *name)
{
    cerr << "Usage: \n"
	<< name << " [options] rootDir mountPoint\n"
	<< "Options:\n"
	<< "  -s      disable multithreaded operation\n"
	<< "  -b      run in background as daemon\n"
	<< "\n"
	<< "Example, to mount at ~/crypt with raw storage in ~/.crypt :\n"
	<< "  encfs -b ~/.crypt ~/crypt\n";
}

const int MaxFuseArgs = 6;
struct EncFs_Args
{
    string rootDir;
    string mountPoint;
    bool isDaemon;
    const char *fuseArgv[MaxFuseArgs];
    int fuseArgc;
};

static char *depathize(char *name)
{
    char *loc = strrchr(name, '/');
    if(loc)
	return loc+1;
    else
	return name;
}

static bool processArgs(int argc, char *argv[], EncFs_Args *out)
{
    out->isDaemon = false;
    out->fuseArgc = 0;

    // pass executable name through
    out->fuseArgv[0] = depathize(argv[0]);
    ++out->fuseArgc;

    enum PathState
    {
	Need_RootDir,
	Need_MountPoint,
	Search_Done
    } state = Need_RootDir;

    // leave a space for mount point, as FUSE expects the mount point before
    // any flags
    out->fuseArgv[1] = 0;
    ++out->fuseArgc;

    for(int i=1; i<argc; ++i)
    {
	if(argv[i][0] == '-')
	{
	    if(argv[i][1] == 's')
	    {
		out->fuseArgv[out->fuseArgc++] = argv[i];
	    } else
	    if(argv[i][1] == 'b')
	    {
		out->isDaemon = true;
	    }else
	    {
		rWarning("Unexpected argument: %s", argv[i]);
		return false;
	    }
	} else
	{
	    // path
	    switch(state)
	    {
	    case Need_RootDir:
		if(!isDirectory( argv[i] ) && !userAllowCreate( argv[i] ))
		{
		    rWarning("Invalid root directory");
		    return false;
		} else
		{
		    out->rootDir = argv[i];
		    if( out->rootDir[ out->rootDir.length()-1 ] != '/' )
			out->rootDir.append( "/" );
		    state = Need_MountPoint;
		}
		break;

	    case Need_MountPoint:
		if(!isDirectory( argv[i] ) && !userAllowCreate( argv[i] ))
		{
		    rWarning("Invalid mount point");
		    return false;
		} else
		{
		    out->mountPoint = argv[i];
		    state = Search_Done;
		}
		break;

	    default:
		// ERROR
		rWarning("Error, unexpected path argument: %s", argv[i]);
		return false;
	    }
	}
    }

    if(out->isDaemon && 
	    (!isAbsolute( out->mountPoint ) ||
	    !isAbsolute( out->rootDir ) ) 
      )
    {
	cerr << "When specifying daemon mode, you can not use relative paths"
	    << endl;
	return false;
    }

    if(state != Search_Done)
    {
	rWarning("Missing one or more arguments\n");
	return false;
    }

    // fill in mount path for fuse
    out->fuseArgv[1] = out->mountPoint.c_str();

    return true;
}

int main(int argc, char *argv[])
{
    // setup cypher
    SSL_load_error_strings();
    SSL_library_init();

    if(argc < 3 || argc >= (1+MaxFuseArgs))
    {
	usage(argv[0]);
	return EXIT_FAILURE;
    }

    // anything that comes from the user should be considered tainted until
    // we've processed it and only allowed through what we support.
    EncFs_Args encfsArgs;
    if(!processArgs(argc, argv, &encfsArgs))
    {
	usage(argv[0]);
	return EXIT_FAILURE;
    }

    rDebug("Root directory: %s", encfsArgs.rootDir.c_str());
    gRootDir = encfsArgs.rootDir;

    gCypher = new Cypher;
    gBlockSize = gCypher->blockSize();

    fuse_operations encfs_oper;
    memset(&encfs_oper, 0, sizeof(fuse_operations));

    encfs_oper.getattr = encfs_getattr;
    encfs_oper.readlink = encfs_readlink;
    encfs_oper.getdir = encfs_getdir;
    encfs_oper.mknod = encfs_mknod;
    encfs_oper.mkdir = encfs_mkdir;
    encfs_oper.unlink = encfs_unlink;
    encfs_oper.rmdir = encfs_rmdir;
    encfs_oper.symlink = encfs_symlink;
    encfs_oper.rename = encfs_rename;
    encfs_oper.link = encfs_link;
    encfs_oper.chmod = encfs_chmod;
    encfs_oper.chown = encfs_chown;
    encfs_oper.truncate = encfs_truncate;
    encfs_oper.utime = encfs_utime;
    encfs_oper.open = encfs_open;
    encfs_oper.read = encfs_read;
    encfs_oper.write = encfs_write;
    encfs_oper.statfs = encfs_statfs;
    encfs_oper.release = encfs_release;
#if defined(HAVE_FUSE_FSYNC)
    encfs_oper.fsync = encfs_fsync;
#endif

    if(initVolumeKey())
    {
	// reset umask now, since we don't want it to interfere with the
	// pass-thru calls..
	umask( 0 );

	if(encfsArgs.isDaemon)
	{
	    RLog_Disable();
	    daemon(0, 0);
	}

	fuse_main(encfsArgs.fuseArgc, const_cast<char**>(encfsArgs.fuseArgv), 
		&encfs_oper);
    }

    delete gCypher;

    return 0;
}
