/*****************************************************************************
 * Author:   Valient Gough <vgough@pobox.com>
 *
 *****************************************************************************
 * Copyright (c) 2002-2003, Valient Gough
 *
 * This library is free software; you can distribute it and/or modify it under
 * the terms of the GNU General Public License (GPL), as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GPL in the file COPYING for more
 * details.
 *
 */

#ifndef _rlog_incl_
#define _rlog_incl_

#ifdef __GNUC__
#  define PRINTF(X,Y) __attribute__ (( __format__ ( __printf__, X, Y)))
#else
#  define PRINTF(X,Y)
#endif

enum RLogType
{
    RLog_Debug,
    RLog_Warning,
    RLog_Error
};

/* An instance of this struct is created for every error message location. 
   By having this per-error data, it makes it possible to enable/disable
   messages on an individual basis.
*/
struct RLogPublisher
{
    /* if node is not null, then this publisher is enabled. */
    struct RLogNode *node;
    int lineNum;
    const char *fileName;
    const char *functionName;
};

/* register a message.  This is used by the rDebug, rWarning, rError macros.
*/
struct RLogPublisher *RLog_Register(struct RLogPublisher **, enum RLogType type,
	const char *fileName, const char *functionName, int lineNum);

/* publish a message, given the publisher information, and a printf style
 * message to publish.  The RLogPublisher struct must have already been
 * registered with RLog_Register.
 *
 *  This function shouldn't be used directly, instead use one of the macros,
 *  rDebug, rWarning, rError.
 */
void RLog_Publish( struct RLogPublisher **, const char *format, ... ) 
	   PRINTF(2,3) ;



/*
    Turn off all message logging.. 
    This is a hack to support daemon operation for the time being.  We should
    probably log anything important to syslog..
*/
void RLog_Disable();

#define CONCAT2(A,B) A##B
#define CONCAT(A,B) CONCAT2(A,B)

#define UNIQUEVAR(PREFIX) CONCAT(PREFIX, __LINE__)


#ifdef __cplusplus
#define _rMessageDef(ID, TYPE) \
  static struct RLogPublisher *ID = \
     RLog_Register( &ID, TYPE, __FILE__, __func__, __LINE__ )
#else
#define _rMessageDef(ID, TYPE) \
  static struct RLogPublisher *ID = 0; \
  if(!ID) \
     ID = RLog_Register( &ID, TYPE, __FILE__, __func__, __LINE__ )
#endif

#define _rMessageCall(ID, ...) \
  if(ID->node) RLog_Publish( &ID, __VA_ARGS__ )

#define _rMessage(ID, TYPE, ... ) \
  do { _rMessageDef(ID, TYPE); _rMessageCall(ID, __VA_ARGS__ ); } while(0)

/*
    Output a debug message - not meant so much for end users as for developers.

    Format is ala printf, eg:
    rDebug("hi %s, how about a nice game of %s?", name, gameType);

    When using a recent GNU compiler, it should automatically detect format
    string / argument mismatch just like it would with printf.
*/
#define rDebug(...) \
  _rMessage( UNIQUEVAR(_rLog_), RLog_Debug, __VA_ARGS__ )

/*
    Output a warning message - meant to indicate that something doesn't seem
    right.
*/
#define rWarning(...) \
  _rMessage( UNIQUEVAR(_rLog_), RLog_Warning, __VA_ARGS__ )

/*
    An error indicates that something has definately gone wrong.
*/
#define rError(...) \
  _rMessage( UNIQUEVAR(_rLog_), RLog_Error, __VA_ARGS__ )

/*
    Assert error condition.  Displays error message if condition does not
    evaluate to TRUE.

    Note that execution continues, although you should not depend on this
    behavior..
*/
#define rAssert( cond ) \
    if( !(cond) ) \
        rError( "Assert failed: " #cond ) \

#endif

